"""loader
flattened XML specific

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022

distribued with GPL-2 or later license

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
"""
from json import dumps
from os.path import isfile, basename

from .i18n import _
from .annotator import CONVERT_OPTION
from .objspace import RootRougailObject
from .error import DictConsistencyError
from .utils import normalize_family


class BaseElt:  # pylint: disable=R0903
    """Base element
    """
    path = '.'


def sorted_func_name(func_name):
    s_func_name = func_name.split('/')
    s_func_name.reverse()
    return '/'.join(s_func_name)


class TiramisuReflector:
    """Convert object to tiramisu representation
    """
    def __init__(self,
                 objectspace,
                 funcs_paths,
                 internal_functions,
                 cfg,
                 ):
        self.cfg = cfg
        self.text = {'header': [],
                     'option': [],
                     'optiondescription': [],
                     }
        if funcs_paths:
            if self.cfg['export_with_import']:
                self.text['header'].extend(["from importlib.machinery import SourceFileLoader as _SourceFileLoader",
                                            "from importlib.util import spec_from_loader as _spec_from_loader, module_from_spec as _module_from_spec",
                                            "class func:",
                                            "    pass",
                                            "",
                                            "def _load_functions(path):",
                                            "    global _SourceFileLoader, _spec_from_loader, _module_from_spec, func",
                                            "    loader = _SourceFileLoader('func', path)",
                                            "    spec = _spec_from_loader(loader.name, loader)",
                                            "    func_ = _module_from_spec(spec)",
                                            "    loader.exec_module(func_)",
                                            "    for function in dir(func_):",
                                            "        if function.startswith('_'):",
                                            "            continue",
                                            "        setattr(func, function, getattr(func_, function))",
                                            ])
            for funcs_path in sorted(funcs_paths, key=sorted_func_name):
                if not isfile(funcs_path):
                    continue
                self.text['header'].append(f"_load_functions('{funcs_path}')")
        if self.cfg['export_with_import']:
            if internal_functions:
                for func in internal_functions:
                    self.text['header'].append(f"setattr(func, '{func}', {func})")
            self.text['header'].extend(["try:",
                                        "    from tiramisu3 import *",
                                        "except:",
                                        "    from tiramisu import *",
                                        ])
        self.objectspace = objectspace
        self.make_tiramisu_objects()
        if self.cfg['export_with_import'] and (self.cfg['force_convert_dyn_option_description'] or self.objectspace.has_dyn_option is True):
            self.text['header'].append("from rougail.tiramisu import ConvertDynOptionDescription")

    def make_tiramisu_objects(self) -> None:
        """make tiramisu objects
        """
        baseelt = BaseElt()
        baseelt.reflector_name = f'option_0{self.objectspace.rougailconfig["suffix"]}'
        self.set_name(baseelt)
        basefamily = Family(baseelt,
                            self.text,
                            self.objectspace,
                            )
        if not self.objectspace.paths.has_path_prefix():
            for elt in self.reorder_family(self.objectspace.space):
                self.populate_family(basefamily,
                                     elt,
                                     )
            basefamily.populate_informations()
            basefamily.elt.information = self.objectspace.paths.get_providers_path()
            basefamily.elt.information.update(self.objectspace.paths.get_suppliers_path())
        else:
            path_prefixes = self.objectspace.paths.get_path_prefixes()
            for path_prefix in path_prefixes:
                space = self.objectspace.space.variables[path_prefix]
                self.set_name(space)
                baseprefix = Family(space,
                                    self.text,
                                    self.objectspace,
                                    )
                basefamily.add(baseprefix)
                for elt in self.reorder_family(space):
                    self.populate_family(baseprefix,
                                         elt,
                                         )
                baseprefix.populate_informations()
                baseprefix.elt.information = self.objectspace.paths.get_providers_path(path_prefix)
                baseprefix.elt.information.update(self.objectspace.paths.get_suppliers_path(path_prefix))
        baseelt.name = normalize_family(self.cfg['base_option_name'])
        baseelt.doc = self.cfg['base_option_name']
        baseelt.reflector_object.get([], baseelt.doc, 'base')  # pylint: disable=E1101

    def reorder_family(self, space):
        """variable_namespace family has to be loaded before any other family
        because `extra` family could use `variable_namespace` variables.
        """
        if hasattr(space, 'variables'):
            variable_namespace = self.objectspace.rougailconfig['variable_namespace']
            if variable_namespace in space.variables:
                yield space.variables[variable_namespace]
            for elt, value in space.variables.items():
                if elt != self.objectspace.rougailconfig['variable_namespace']:
                    yield value
        if hasattr(space, 'services'):
            yield space.services

    def populate_family(self,
                        parent_family,
                        elt,
                        ):
        """Populate family
        """
        self.set_name(elt)
        family = Family(elt,
                        self.text,
                        self.objectspace,
                        )
        parent_family.add(family)
        for children in vars(elt).values():
            if isinstance(children, self.objectspace.family):
                self.populate_family(family,
                                     children,
                                     )
                continue
            if isinstance(children, dict):
                children = list(children.values())
            if isinstance(children, list):
                for child in children:
                    if isinstance(child, self.objectspace.property_) or \
                            not isinstance(child, RootRougailObject):
                        continue
                    if isinstance(child, self.objectspace.variable):
                        self.set_name(child)
                        family.add(Variable(child,
                                            self.text,
                                            self.objectspace,
                                            ))
                    else:
                        self.populate_family(family,
                                             child,
                                             )

    def set_name(self,
                 elt,
                 ):
        """Set name
        """
        if not hasattr(elt, 'reflector_name'):
            self.objectspace.paths.set_name(elt, 'optiondescription_')
        return elt.reflector_name

    def get_text(self):
        """Get text
        """
        return '\n'.join(self.text['header'] + self.text['option'] + self.text['optiondescription'])


class Common:
    """Common function for variable and family
    """
    def __init__(self,
                 elt,
                 text,
                 objectspace,
                 ):
        self.elt = elt
        self.option_name = None
        self.text = text
        self.objectspace = objectspace
        self.elt.reflector_object = self
        self.object_type = None

    def get(self, calls, parent_name, typ):
        """Get tiramisu's object
        """
        self_calls = calls.copy()
        if self.elt.path in self_calls:
            msg = f'"{self.elt.path}" will make an infinite loop'
            raise DictConsistencyError(msg, 80, self.elt.xmlfiles)
        self_calls.append(self.elt.path)
        self.calls = self_calls
        if self.option_name is None:
            self.option_name = self.elt.reflector_name
            self.populate_attrib()
            self.populate_informations()
        return self.option_name

    def populate_attrib(self):
        """Populate attributes
        """
        keys = {'name': self.convert_str(self.elt.name)}
        if hasattr(self.elt, 'doc'):
            keys['doc'] = self.convert_str(self.elt.doc)
        self._populate_attrib(keys)
        if hasattr(self.elt, 'properties'):
            keys['properties'] = self.properties_to_string(self.elt.properties)
        attrib = ', '.join([f'{key}={value}' for key, value in keys.items()])
        if self.__class__.__name__ == 'Family':
            #pouet
            name = 'option'
            #name = 'optiondescription'
        else:
            name = 'option'
        self.text[name].append(f'{self.option_name} = {self.object_type}({attrib})')

    def _populate_attrib(self,
                         keys: dict,
                         ) -> None:  # pragma: no cover
        raise NotImplementedError()

    @staticmethod
    def convert_str(value):
        """convert string
        """
        return dumps(value, ensure_ascii=False)

    def properties_to_string(self,
                             values: list,
                             ) -> None:
        """Change properties to string
        """
        properties = [self.convert_str(property_) for property_ in values
                      if isinstance(property_, str)]
        calc_properties = [self.calc_properties(property_) for property_ in values \
                if isinstance(property_, self.objectspace.property_)]
        return 'frozenset({' + ', '.join(sorted(properties) + calc_properties) + '})'

    def calc_properties(self,
                        child,
                        ) -> str:
        """Populate properties
        """
        option_name = child.source.reflector_object.get(self.calls, self.elt.path, 'property')
        kwargs = (f"'condition': ParamOption({option_name}, todict=True, notraisepropertyerror=True), "
                  f"'expected': {self.populate_param(child.expected)}")
        if child.inverse:
            kwargs += ", 'reverse_condition': ParamValue(True)"
        return (f"Calculation(func.calc_value, Params(ParamValue('{child.name}'), "
                f"kwargs={{{kwargs}}}))")

    def populate_informations(self):
        """Populate Tiramisu's informations
        """
        if not hasattr(self.elt, 'information'):
            return
        if isinstance(self.elt.information, dict):
            informations = self.elt.information
        else:
            informations = vars(self.elt.information)
        for key, value in informations.items():
            if key == 'xmlfiles':
                continue
            if isinstance(value, str):
                value = self.convert_str(value)
            self.text['option'].append(f"{self.option_name}.impl_set_information('{key}', {value})")

    def populate_param(self,
                       param,
                       ):
        """Populate variable parameters
        """
        if param.type in ['number', 'boolean', 'nil', 'string', 'port', 'choice', 'space']:
            value = param.text
            if param.type == 'string' and value is not None:
                value = self.convert_str(value)
            return f'ParamValue({value})'
        if param.type in ['variable_name', 'variable']:
            return self.build_option_param(param)
        if param.type == 'information':
            if hasattr(self.elt, 'multi') and self.elt.multi:
                default = []
            else:
                default = None
            return f'ParamInformation("{param.text}", {default})'
        if param.type == 'target_information':
            return f'ParamSelfInformation("{param.text}", None)'
        if param.type == 'suffix':
            return 'ParamSuffix()'
        if param.type == 'index':
            return 'ParamIndex()'
        raise Exception(_(f'unknown type {param.type}'))  # pragma: no cover

    def build_option_param(self,
                           param,
                           ) -> str:
        """build variable parameters
        """
        if param.type == 'variable':
            option_name = param.text.reflector_object.get(self.calls, self.elt.path, 'param')
        else:
            option_name = param.text
        params = [f'{option_name}']
        if hasattr(param, 'suffix'):
            param_type = 'ParamDynOption'
            family = param.family.reflector_object.get(self.calls, self.elt.path, 'suffix')
            params.extend([f"'{param.suffix}'", f'{family}'])
        else:
            param_type = 'ParamOption'
        if not param.propertyerror:
            params.append('notraisepropertyerror=True')
        return "{}({})".format(param_type, ', '.join(params))


class Variable(Common):
    """Manage variable
    """
    def __init__(self,
                 elt,
                 text,
                 objectspace,
                 ):
        super().__init__(elt, text, objectspace)
        self.object_type = CONVERT_OPTION[elt.type]['opttype']

    def _populate_attrib(self,
                         keys: dict,
                         ):
        if hasattr(self.elt, 'opt'):
            keys['opt'] = self.elt.opt.reflector_object.get(self.calls, self.elt.path, 'opt')
        if hasattr(self.elt, 'choice'):
            values = self.elt.choice
            if values[0].type == 'variable':
                value = values[0].name.reflector_object.get(self.calls, self.elt.path, 'choice')
                keys['values'] = f"Calculation(func.calc_value, Params((ParamOption({value}))))"
            elif values[0].type == 'function':
                keys['values'] = self.calculation_value(values[0], [])
            else:
                keys['values'] = str(tuple([val.name for val in values]))
        if hasattr(self.elt, 'multi') and self.elt.multi:
            keys['multi'] = self.elt.multi
        for key in ['default', 'default_multi']:
            if hasattr(self.elt, key) and getattr(self.elt, key) is not None:
                value = getattr(self.elt, key)
                if isinstance(value, str):
                    value = self.convert_str(value)
                elif isinstance(value, self.objectspace.value):
                    value = self.calculation_value(value, [], calc_multi=value.calc_multi)
                keys[key] = value
        if hasattr(self.elt, 'validators'):
            keys['validators'] = '[' + ', '.join([self.calculation_value(val,
                                    ['ParamSelfOption(whole=False)']) for val in self.elt.validators]) + ']'
        for key in ['min_number', 'max_number']:
            if hasattr(self.elt, key):
                keys[key] = getattr(self.elt, key)
        for key, value in CONVERT_OPTION[self.elt.type].get('initkwargs', {}).items():
            if isinstance(value, str):
                value = f"'{value}'"
            keys[key] = value

    def calculation_value(self,
                          child,
                          args,
                          calc_multi=False,
                          ) -> str:
        """Generate calculated value
        """
        kwargs = []
        # has parameters
        function = child.name
        new_args = []
        if hasattr(child, 'param'):
            for param in child.param:
                value = self.populate_param(param)
                if not hasattr(param, 'name'):
                    new_args.append(str(value))
                else:
                    kwargs.append(f"'{param.name}': " + value)
        if function == 'valid_network_netmask':
            new_args.extend(args)
        else:
            args.extend(new_args)
            new_args = args
        ret = f'Calculation(func.{function}, Params((' + ', '.join(new_args) + ')'
        if kwargs:
            ret += ', kwargs={' + ', '.join(kwargs) + '}'
        ret += ')'
        if hasattr(child, 'warnings_only'):
            ret += f', warnings_only={child.warnings_only}'
        ret = ret + ')'
        if calc_multi:
            ret = '[' + ret + ']'
        return ret


class Family(Common):
    """Manage family
    """
    def __init__(self,
                 elt,
                 text,
                 objectspace,
                 ):
        super().__init__(elt, text, objectspace)
        if hasattr(self.elt, 'suffixes'):
            self.objectspace.has_dyn_option = True
            self.object_type = 'ConvertDynOptionDescription'
        elif hasattr(self.elt, 'leadership') and self.elt.leadership:
            self.object_type = 'Leadership'
        else:
            self.object_type = 'OptionDescription'
        self.children = []

    def add(self, child):
        """Add a child
        """
        self.children.append(child)

    def _populate_attrib(self,
                         keys: list,
                         ) -> None:
        if hasattr(self.elt, 'suffixes'):
            dyn = self.elt.suffixes.reflector_object.get(self.calls, self.elt.path, 'suffixes')
            keys['suffixes'] = f"Calculation(func.calc_value, Params((ParamOption({dyn}, notraisepropertyerror=True))))"
        keys['children'] = '[' + ', '.join([child.get(self.calls, self.elt.path, 'child') for child in self.children]) + ']'
